schema_metas = {
    'Jina::Metas': {
        'description': 'The meta config of the Executor',
        'type': 'object',
        'required': [],
        'additionalProperties': False,
        'properties': {
            'is_trained': {
                'description': 'Indicates if the executor is trained or not. '
                'If not, then methods decorated by `@required_train` can not be executed.',
                'type': 'boolean',
                'default': False,
            },
            'is_updated': {
                'description': 'Indicates if the executor is updated or changed since last save. '
                'If not, then save() will do nothing. A forced save is possible to use `touch()` before `save()`',
                'type': 'boolean',
                'default': False,
            },
            'batch_size': {
                'description': 'The size of each batch, methods decorated by `@batching` will respect this. '
                'Useful when incoming data is too large to fit into (GPU) memory.',
                'type': 'number',
                'default': None,
            },
            'workspace': {
                'description': '''
The working directory, for persisting the artifacts of the executor. An artifact is a file or collection of files used during a workflow run.

By default it is not set, if you expect your executor to be persisted or to persist any data, remember to set it to the desired value.

When a BaseExecutor is a component of a `CompoundExecutor`, its workspace value will be overridden by the workspace coming from the `CompoundExecutor` unless a particular workspace value is set for the component BaseExecutor.                
                ''',
                'type': 'string',
                'default': None,
            },
            'name': {
                'description': 'The name of the executor.',
                'type': 'string',
                'default': None,
            },
            'description': {
                'description': 'The description of this executor. It will be used in automatics docs UI',
                'type': 'string',
                'default': None,
            },
            'read_only': {
                'description': 'Do not allow the Pod to modify the Executor, save calls will be ignored. '
                'If set to true no serialization of the Executor',
                'type': 'boolean',
                'default': False,
            },
            'on_gpu': {
                'description': 'If the executor is set to run on GPU.',
                'type': 'boolean',
                'default': False,
            },
            'py_modules': {
                'type': 'array',
                'default': None,
                'items': {'type': 'string', 'minItems': 1, 'uniqueItems': True},
                'description': '''
The external python module paths. it is useful when you want to load external python modules
using :func:`BaseExecutor.load_config` from a YAML file. If a relative path is given then the root path is set to
the path of the current YAML file.

Example of ``py_module`` usage:

1. This is a valid structure and it is RECOMMENDED:
    - "my_cust_module" is a python module
    - all core logic of your customized executor goes to ``__init__.py``
    - to import ``foo.py``, you can use relative import, e.g. ``from .foo import bar``
    - ``helper.py`` needs to be put BEFORE `__init__.py` in YAML ``py_modules``

This is also the structure given by ``jina hub new`` CLI.

    .. highlight:: text
    .. code-block:: text

         my_cust_module
           |- __init__.py
           |- helper.py
           |- config.yml
                |- py_modules
                       |- helper.py
                       |- __init__.py

2. This is a valid structure but not recommended:
    - "my_cust_module" is not a python module (lack of __init__.py under the root)
    - to import ``foo.py``, you must to use ``from jinahub.foo import bar``
    - ``jinahub`` is a common namespace for all plugin-modules, not changeable.
    - ``helper.py`` needs to be put BEFORE `my_cust.py` in YAML ``py_modules``

    .. highlight:: text
    .. code-block:: text

         my_cust_module
           |- my_cust.py
           |- helper.py
           |- config.yml
                |- py_modules
                       |- helper.py
                       |- my_cust.py                
                ''',
            },
        },
    }
}



import os
import numpy as np
import pickle

from umap import UMAP

target_output_dim = 2
batch_size = 10
input_dim = 28
model_path = 'pre-trained.model'

train_data = np.random.rand(batch_size, input_dim)

model = UMAP(n_components=target_output_dim, random_state=42)
model.fit(train_data)
pickle.dump(model, open(model_path, 'wb'))



import inspect
from typing import Dict, List, Union

from .. import BaseExecutor
from ...helper import typename


class BaseSegmenter(BaseExecutor):
    """:class:`BaseSegmenter` works on doc-level,
    it chunks Documents into set of Chunks
    :param args: Variable length arguments
    :param kwargs: Variable length keyword arguments
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.required_keys = [
            k for k in inspect.getfullargspec(self.segment).args if k != 'self'
        ]
        if not self.required_keys:
            self.required_keys = [
                k
                for k in inspect.getfullargspec(inspect.unwrap(self.segment)).args
                if k != 'self'
            ]
        if not self.required_keys:
            self.logger.warning(
                f'{typename(self)} works on keys, but no keys are specified'
            )

    def segment(self, *args, **kwargs) -> Union[List[List[Dict]], List[Dict]]:
        """
        :param args: Variable length arguments
        :param kwargs: Variable length keyword arguments
        """
        raise NotImplementedError



__copyright__ = "Copyright (c) 2021 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from typing import Dict, Union, Optional

import numpy as np
from jina.executors.rankers import Chunk2DocRanker


class BiMatchRanker(Chunk2DocRanker):
    """The :class:`BiMatchRanker` counts the best chunk-hit from both query and doc perspective.

    `BiMatchRanker` tries to assign a score for a `match` given a `query` based on the matches found for the `query`
    chunks on the `match` chunks.

    The situation can be seen with an hypothetical example:

    `Documents in the Index`:
        - Granularity 0 - Doc 1 - `BiMatchRanker is a Hub Executor. computes relevance scores. Found in Jina, the neural search framework`
                        - Granularity 1 - Doc 11 - `BiMatchRanker is a Ranking Jina Hub Executor`
                                        - Doc 12 - `It is a powerfull tool.`
                                        - Doc 13 - `Found in Jina`
                                        - Doc 14 - `the neural search framework`
        - Granularity 0 - Doc 2 - `A Ranker executor gives relevance Scores. Rankers are found in the Hub, and built automatically`
                        - Granularity 1 - Doc 21 - `A Ranker executor gives relevance Scores`
                        - Granularity 1 - Doc 22 - `Rankers are found in the Hub`
                        - Granularity 1 - Doc 23 - `and built automatically`

    `Query`
        - Granularity 0 - 'A Hub Executor. Ranking on Relevance scores. Giving automatically relevant documents from the Hub'
                        - Granularity 1 - `A Hub Executor`
                                        - `Ranking on Relevance scores`
                                        - `Giving automatically relevant documents from the Hub`

    `Semantic similarity matches` (TOP_K 3)
        Query chunk 1 -> `A Hub Executor` -> Doc11, Doc22, Doc21 (Children of `Doc1` and `Doc2`)
        Query chunk 2 -> `Ranking on Relevance scores` -> Doc21, Doc11, Doc22 (Children of `Doc2` and `Doc1`)
        Query chunk 3 -> `Scoring relevant documents`  -> Doc21, Doc22, Doc23 (Children of `Doc2`)

    In order to compute this relevance score, it tries to take into account two perspectives.

        - Match perspective: How many chunks of a match are part of the set of matches (hit) of a query (and its chunks).
            In the example:
                - `Doc1` has 4 chunks, but only 1 (hit) (Doc11) is found in the matches of the query chunks.
                - `Doc2` has 3 chunks and 2 of them (hit) are found in the matches (Doc21, Doc22).

        - Query perspective: Of all the chunks in the query, how many of them are matches by any of the chunks of each match document.
            In the example:
                - `Query` for `Doc1`: Query has 3 chunks, but only 2 of them are matched by chunks of match `Doc1` (Doc11)
                - `Query` for `Doc2`: Query has 3 chunks, and the 3 of them are matched by chunks of match `Doc2`

    So for every `Document`, a score is computed adding some penalty to the `missing` in both perspectives.

    :param d_miss: Cost associated to a miss chunk
    :param args:  Additional positional arguments
    :param kwargs: Additional keyword arguments

    .. warning:: Here we suppose that the smaller chunk score means the more similar.
    """

    def __init__(self, d_miss: Optional[Union[int, float]] = None, *args, **kwargs):
        super().__init__(query_required_keys=('length', ), match_required_keys=('length', ), *args, **kwargs)
        self.d_miss = d_miss or 2000

    def score(self, match_idx: 'np.ndarray', query_chunk_meta: Dict, match_chunk_meta: Dict, *args, **kwargs):
        # match_idx [parent_match_id, match_id, query_id, score]
        # all matches have the same parent_id (the parent_id of the matches)
        s1 = self._directional_score(match_idx, match_chunk_meta, col=self.COL_DOC_CHUNK_ID)
        s2 = self._directional_score(match_idx, query_chunk_meta, col=self.COL_QUERY_CHUNK_ID)
        return (s1 + s2) / 2.

    def _group_by(self, match_idx, col_name):
        # sort by ``col
        _sorted_m = np.sort(match_idx, order=col_name)
        _, _doc_counts = np.unique(_sorted_m[col_name], return_counts=True)
        # group by ``col``
        return np.split(_sorted_m, np.cumsum(_doc_counts))[:-1]

    def _directional_score(self, g: Dict, chunk_meta: Dict, col: str):
        # g [parent_match_id, match_id, query_id, score]
        # col = self.COL_DOC_CHUNK_ID, from matched_chunk aspect
        # col = self.COL_QUERY_CHUNK_ID, from query chunk aspect
        # group by "match_id" or "query_chunk_id". So here groups have in common 1st (match_parent_id) and `n` column
        _groups = self._group_by(g, col)
        # take the best match from each group
        _groups_best = np.stack([np.sort(gg, order='score')[0] for gg in _groups])
        # doc total length
        # how many chunks in the document (match or query)
        _c = chunk_meta[_groups_best[0][col]]['length']
        # how many chunks hit
        _h = _groups_best.shape[0]
        # hit distance
        sum_d_hit = np.sum(_groups_best[self.COL_SCORE])
        # all hit => 0, all_miss => 1
        return 1 - (sum_d_hit + self.d_miss * (_c - _h)) / (self.d_miss * _c)



from typing import List

import numpy as np

from . import BaseSparseNdArray
from ....proto import jina_pb2

if False:
    import scipy.sparse

__all__ = ['SparseNdArray']


class SparseNdArray(BaseSparseNdArray):
    """
    Scipy powered sparse ndarray.

    .. warning::
        scipy only supports ndim=2

    .. seealso::
        https://docs.scipy.org/doc/scipy/reference/sparse.html

    :param proto: the protobuf message, when not given then create a new one via :meth:`get_null_proto`
    :param sp_format: the sparse format of the scipy matrix. one of 'coo', 'bsr', 'csc', 'csr'.
    """

    def __init__(
        self,
        proto: 'jina_pb2.SparseNdArrayProto' = None,
        sp_format: str = 'coo',
        *args,
        **kwargs,
    ):
        """Set constructor method."""
        import scipy.sparse

        super().__init__(proto, *args, **kwargs)
        support_fmt = {'coo', 'bsr', 'csc', 'csr'}
        if sp_format in support_fmt:
            self.spmat_fn = getattr(scipy.sparse, f'{sp_format}_matrix')
        else:
            raise ValueError(
                f'{sp_format} sparse matrix is not supported, please choose one of those: {support_fmt}'
            )

    def sparse_constructor(
        self, indices: 'np.ndarray', values: 'np.ndarray', shape: List[int]
    ) -> 'scipy.sparse.spmatrix':
        """
        Sparse NdArray constructor for scipy.sparse.spmatrix.

        :param indices: the indices of the sparse array
        :param values: the values of the sparse array
        :param shape: the shape of the sparse array
        :return: SparseTensor
        """
        if indices.shape[-1] != 2:
            raise ValueError(
                f'scipy backend only supports ndim=2 sparse matrix, given {indices.shape}'
            )
        return self.spmat_fn((values, indices.T), shape=shape)

    def sparse_parser(self, value: 'scipy.sparse.spmatrix'):
        """
        Parse a scipy.sparse.spmatrix to indices, values and shape.

        :param value: the scipy.sparse.spmatrix.
        :return: a Dict with three entries {'indices': ..., 'values':..., 'shape':...}
        """
        v = value.tocoo()
        return {
            'indices': np.stack([v.row, v.col], axis=1),
            'values': v.data,
            'shape': v.shape,
        }



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from typing import Optional

from . import FlatRecursiveMixin, BaseExecutableDriver
from ..excepts import LengthMismatchException

if False:
    from .. import DocumentSet


class CraftDriver(FlatRecursiveMixin, BaseExecutableDriver):
    """Drivers inherited from this Driver will bind :meth:`craft` by default """

    def __init__(
        self, executor: Optional[str] = None, method: str = 'craft', *args, **kwargs
    ):
        super().__init__(executor, method, *args, **kwargs)

    def _apply_all(self, docs: 'DocumentSet', *args, **kwargs):
        if not docs:
            self.logger.warning(f'an empty DocumentSet {docs}')
            return

        contents, docs_pts = docs.extract_docs(*self.exec.required_keys)

        if not docs_pts:
            self.logger.warning(f'no Document is extracted {docs}')
            return

        if len(self.exec.required_keys) > 1:
            craft_dicts = self.exec_fn(*contents)
        else:
            craft_dicts = self.exec_fn(contents)

        if len(docs_pts) != len(craft_dicts):
            msg = (
                f'mismatched {len(docs_pts)} docs from level {docs_pts[0].granularity} '
                f'and length of returned crafted documents: {len(craft_dicts)}, the length must be the same'
            )
            self.logger.error(msg)
            raise LengthMismatchException(msg)

        for doc, crafted in zip(docs_pts, craft_dicts):
            doc.set_attrs(**crafted)



"""Argparser module for the check functions"""
from .base import set_base_parser


def set_check_parser(parser=None):
    """Set the `check` parser

    :param parser: an optional existing parser to build upon
    :return: parser
    """
    if not parser:
        parser = set_base_parser()

    parser.add_argument(
        '--summary-exec',
        type=str,
        help='The markdown file path for all executors summary',
    )
    parser.add_argument(
        '--summary-driver',
        type=str,
        help='The markdown file path for all drivers summary',
    )
    return parser



from typing import Dict, List, Tuple
from collections import defaultdict, namedtuple

import numpy as np

from ....executors.rankers import Chunk2DocRanker
from ....types.document import Document
from ....types.score import NamedScore

from .. import BaseRankDriver

if False:
    from ....types.sets import DocumentSet

COL_STR_TYPE = 'U64'  #: the ID column data type for score matrix


class BaseAggregateMatchesRankerDriver(BaseRankDriver):
    """Drivers inherited from this Driver focus on aggregating scores from `chunks` to its `parents`.

    :param keep_source_matches_as_chunks: A flag to indicate if the driver must return the old matches of the query or its chunks
            (at a greater granularity level (k + 1)) as the chunks of the new computed `matches` (at granularity level k)
            Set it to `True` when keeping track of the chunks that lead to a retrieved result.
    :param args: additional positional arguments which are just used for the parent initialization
    :param kwargs: additional key value arguments which are just used for the parent initialization

    .. note::
        When set `keep_source_matches_as_chunks=True`, the chunks of the match contains **ONLY** the chunks leading
        to the match rather than **ALL** the chunks of the match."""

    def __init__(self, keep_source_matches_as_chunks: bool = False, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.keep_source_matches_as_chunks = keep_source_matches_as_chunks

    QueryMatchInfo = namedtuple(
        'QueryMatchInfo', 'match_parent_id match_id query_id score'
    )

    def _extract_query_match_info(self, match: Document, query: Document):
        return self.QueryMatchInfo(
            match_parent_id=match.parent_id,
            match_id=match.id,
            query_id=query.id,
            score=match.score.value,
        )

    def _insert_query_matches(
        self,
        query: Document,
        parent_id_chunk_id_map: dict,
        chunk_matches_by_id: dict,
        docs_scores: 'np.ndarray',
    ):
        """
        :param query: the query Document where the resulting matches will be inserted
        :param parent_id_chunk_id_map: a map with parent_id as key and list of previous matches ids as values
        :param chunk_matches_by_id: the previous matches of the query (at a higher granularity) grouped by the new map (by its parent)
        :param docs_scores: An `np.ndarray` resulting from the ranker executor with the `scores` of the new matches
        """

        op_name = self.exec.__class__.__name__
        for doc_id, score in docs_scores:
            m = Document(id=doc_id)
            m.score = NamedScore(op_name=op_name, value=score)
            if self.keep_source_matches_as_chunks:
                for match_chunk_id in parent_id_chunk_id_map[doc_id]:
                    m.chunks.append(chunk_matches_by_id[match_chunk_id])
            query.matches.append(m)

    @staticmethod
    def _group_by(match_idx, col_name):
        """
        Create an list of numpy arrays with the same ``col_name`` in each position of the list

        :param match_idx: Numpy array of Tuples with document id and score
        :param col_name:  Column name in the structured numpy array of Tuples

        :return: List of numpy arrays with the same ``doc_id`` in each position of the list
        :rtype: np.ndarray.
        """
        _sorted_m = np.sort(match_idx, order=col_name)
        n_elements = len(_sorted_m[col_name])
        list_numpy_arrays = []
        prev_val = _sorted_m[col_name][0]
        prev_index = 0
        for i, current_val in enumerate(_sorted_m[col_name]):
            if current_val != prev_val:
                list_numpy_arrays.append(_sorted_m[prev_index:i])
                prev_index = i
                prev_val = current_val
        list_numpy_arrays.append(_sorted_m[prev_index:])
        return list_numpy_arrays

    @staticmethod
    def _sort_doc_by_score(r):
        """
        Sort a numpy array  of dtype (``doc_id``, ``score``) by the ``score``.

        :param r: Numpy array of Tuples with document id and score
        :type r: np.ndarray[Tuple[np.str_, np.float64]]
        """
        r[::-1].sort(order=Chunk2DocRanker.COL_SCORE)

    def _score(
        self, match_idx: 'np.ndarray', query_chunk_meta: Dict, match_chunk_meta: Dict
    ) -> 'np.ndarray':
        """
        Translate the chunk-level top-k results into doc-level top-k results. Some score functions may leverage the
        meta information of the query, hence the meta info of the query chunks and matched chunks are given
        as arguments.

        :param match_idx: A [N x 4] numpy ``ndarray``, column-wise:
                - ``match_idx[:, 0]``: ``doc_id`` of the matched chunks, integer
                - ``match_idx[:, 1]``: ``chunk_id`` of the matched chunks, integer
                - ``match_idx[:, 2]``: ``chunk_id`` of the query chunks, integer
                - ``match_idx[:, 3]``: distance/metric/score between the query and matched chunks, float
        :type match_idx: np.ndarray.
        :param query_chunk_meta: The meta information of the query chunks, where the key is query chunks' ``chunk_id``,
            the value is extracted by the ``query_required_keys``.
        :param match_chunk_meta: The meta information of the matched chunks, where the key is matched chunks'
            ``chunk_id``, the value is extracted by the ``match_required_keys``.
        :return: A [N x 2] numpy ``ndarray``, where the first column is the matched documents' ``doc_id`` (integer)
                the second column is the score/distance/metric between the matched doc and the query doc (float).
        :rtype: np.ndarray.
        """
        _groups = self._group_by(match_idx, Chunk2DocRanker.COL_PARENT_ID)
        n_groups = len(_groups)
        res = np.empty(
            (n_groups,),
            dtype=[
                (Chunk2DocRanker.COL_PARENT_ID, COL_STR_TYPE),
                (Chunk2DocRanker.COL_SCORE, np.float64),
            ],
        )

        for i, _g in enumerate(_groups):
            res[i] = (
                _g[Chunk2DocRanker.COL_PARENT_ID][0],
                self.exec_fn(_g, query_chunk_meta, match_chunk_meta),
            )

        self._sort_doc_by_score(res)
        return res


class Chunk2DocRankDriver(BaseAggregateMatchesRankerDriver):
    """Extract matches score from chunks and use the executor to compute the rank and assign the resulting matches to the
    level above.

    Input-Output ::
        Input:
        document: {granularity: k-1}
                |- chunks: {granularity: k}
                |    |- matches: {granularity: k}
                |
                |- chunks: {granularity: k}
                    |- matches: {granularity: k}
        Output:
        document: {granularity: k-1}
                |- chunks: {granularity: k}
                |    |- matches: {granularity: k}
                |
                |- chunks: {granularity: k}
                |    |- matches: {granularity: k}
                |
                |-matches: {granularity: k-1} (Ranked according to Ranker Executor)
    """

    def __init__(self, traversal_paths: Tuple[str] = ('r',), *args, **kwargs):
        super().__init__(traversal_paths=traversal_paths, *args, **kwargs)

    def _apply_all(self, docs: 'DocumentSet', *args, **kwargs) -> None:
        """
        :param docs: the doc which gets bubbled up matches
        :param args: not used (kept to maintain interface)
        :param kwargs: not used (kept to maintain interface)
        """
        for doc in docs:
            chunks = doc.chunks
            match_idx = []  # type: List[Tuple[str, str, str, float]]
            query_meta = {}  # type: Dict[str, Dict]
            match_meta = {}  # type: Dict[str, Dict]
            parent_id_chunk_id_map = defaultdict(list)
            matches_by_id = defaultdict(Document)
            for chunk in chunks:
                query_meta[chunk.id] = (
                    chunk.get_attrs(*self._exec_query_keys)
                    if self._exec_query_keys
                    else None
                )
                for match in chunk.matches:
                    match_info = self._extract_query_match_info(
                        match=match, query=chunk
                    )
                    match_idx.append(match_info)
                    match_meta[match.id] = (
                        match.get_attrs(*self._exec_match_keys)
                        if self._exec_match_keys
                        else None
                    )
                    parent_id_chunk_id_map[match.parent_id].append(match.id)
                    matches_by_id[match.id] = match

            if match_idx:
                match_idx = np.array(
                    match_idx,
                    dtype=[
                        (Chunk2DocRanker.COL_PARENT_ID, COL_STR_TYPE),
                        (Chunk2DocRanker.COL_DOC_CHUNK_ID, COL_STR_TYPE),
                        (Chunk2DocRanker.COL_QUERY_CHUNK_ID, COL_STR_TYPE),
                        (Chunk2DocRanker.COL_SCORE, np.float64),
                    ],
                )

                docs_scores = self._score(match_idx, query_meta, match_meta)

                self._insert_query_matches(
                    query=doc,
                    parent_id_chunk_id_map=parent_id_chunk_id_map,
                    chunk_matches_by_id=matches_by_id,
                    docs_scores=docs_scores,
                )


class AggregateMatches2DocRankDriver(BaseAggregateMatchesRankerDriver):
    """This Driver is intended to take a `document` with matches at a `given granularity > 0`, clear those matches and substitute
    these matches by the documents at a lower granularity level.
    Input-Output ::
        Input:
        document: {granularity: k}
            |- matches: {granularity: k}

        Output:
        document: {granularity: k}
            |- matches: {granularity: k-1} (Sorted according to Ranker Executor)

    Imagine a case where we are querying a system with text documents chunked by sentences. When we query the system,
    we use sentences (chunks) to query it. So at some point we will have:
    `query sentence (documents of granularity 1):
        matches: indexed sentences (documents of level depth 1)`
    `
    But in the output we want to have the full document that better matches the `sentence`.
    `query sentence (documents of granularity 1):
        matches: indexed full documents (documents of granularity 0).
    `
    Using this Driver before querying a Binary Index with full binary document data can be very useful to implement a search system.
    """

    def __init__(self, traversal_paths: Tuple[str] = ('r',), *args, **kwargs):
        super().__init__(traversal_paths=traversal_paths, *args, **kwargs)

    def _apply_all(self, docs: 'DocumentSet', *args, **kwargs) -> None:
        """

        :param docs: the document at granularity ``k``
        :param args: not used (kept to maintain interface)
        :param kwargs: not used (kept to maintain interface)

        .. note::
            - This driver will substitute the ``matches`` of `docs` to the corresponding ``parent documents`` of its current ``matches`` according
            to the executor.
            - Set the ``traversal_paths`` of this driver to identify the documents, which needs to get bubbled up matches.
        """

        for doc in docs:
            matches = doc.matches

            match_idx = []
            query_meta = {}
            match_meta = {}
            parent_id_chunk_id_map = defaultdict(list)
            matches_by_id = defaultdict(Document)

            query_meta[doc.id] = (
                doc.get_attrs(*self._exec_query_keys) if self._exec_query_keys else None
            )

            for match in matches:
                match_info = self._extract_query_match_info(match=match, query=doc)
                match_idx.append(match_info)
                match_meta[match.id] = (
                    match.get_attrs(*self._exec_match_keys)
                    if self._exec_match_keys
                    else None
                )
                parent_id_chunk_id_map[match.parent_id].append(match.id)
                matches_by_id[match.id] = match

            if match_idx:
                match_idx = np.array(
                    match_idx,
                    dtype=[
                        (Chunk2DocRanker.COL_PARENT_ID, COL_STR_TYPE),
                        (Chunk2DocRanker.COL_DOC_CHUNK_ID, COL_STR_TYPE),
                        (Chunk2DocRanker.COL_QUERY_CHUNK_ID, COL_STR_TYPE),
                        (Chunk2DocRanker.COL_SCORE, np.float64),
                    ],
                )

                docs_scores = self._score(match_idx, query_meta, match_meta)
                # This ranker will change the current matches
                doc.ClearField('matches')
                self._insert_query_matches(
                    query=doc,
                    parent_id_chunk_id_map=parent_id_chunk_id_map,
                    chunk_matches_by_id=matches_by_id,
                    docs_scores=docs_scores,
                )



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

import sys

from typing import Iterable, Optional

from .. import QuerySetReader, ContextAwareRecursiveMixin, BaseRecursiveDriver

if False:
    from ...types.sets.document import DocumentSet


class SliceQL(QuerySetReader, ContextAwareRecursiveMixin, BaseRecursiveDriver):
    """Restrict the size of the ``docs`` to ``k`` (given by the request)

    Example::
    - !ReduceAllDriver
        with:
            traversal_paths: ['m']
    - !SortQL
        with:
            reverse: true
            field: 'score__value'
            traversal_paths: ['m']
    - !SliceQL
        with:
            start: 0
            end: 50
            traversal_paths: ['m']

    `SliceQL` will ensure that only the first 50 documents are returned from this `Pod`

    :param start: Zero-based index at which to start extraction.
    :param end:  Zero-based index before which to end extraction.
            slice extracts up to but not including end. For example, take(1,4) extracts
            the second element through the fourth element (elements indexed 1, 2, and 3).
    :param args: additional positional arguments which are just used for the parent initialization
    :param kwargs: additional key value arguments which are just used for the parent initialization
    """

    def __init__(self, start: int, end: Optional[int] = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._start = int(start)
        if end is None:
            self._end = sys.maxsize
        else:
            self._end = int(end)

    def _apply_all(
        self, doc_sequences: Iterable['DocumentSet'], *args, **kwargs
    ) -> None:
        for docs in doc_sequences:
            if self.start <= 0 and (self.end is None or self.end >= len(docs)):
                pass
            else:
                del docs[int(self.end) :]
                del docs[: int(self.start)]



"""Argparser module for Flow"""
from .base import set_base_parser
from .peapods.base import mixin_base_ppr_parser
from ..enums import FlowInspectType


def set_flow_parser(parser=None):
    """Set the parser for the flow

    :param parser: an (optional) initial parser to build upon
    :return: the parser
    """
    if not parser:
        parser = set_base_parser()

    mixin_base_ppr_parser(parser)

    parser.add_argument('--uses', type=str, help='The YAML file represents a flow')
    parser.add_argument(
        '--inspect',
        type=FlowInspectType.from_string,
        choices=list(FlowInspectType),
        default=FlowInspectType.COLLECT,
        help='''
The strategy on those inspect pods in the flow.

If `REMOVE` is given then all inspect pods are removed when building the flow.
''',
    )

    return parser



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

import gzip
import io
import os
from functools import lru_cache
from os import path
from typing import Optional, Iterable, Tuple, Dict

import numpy as np

from . import BaseVectorIndexer
from ..decorators import batching
from ...helper import cached_property
from ...importer import ImportExtensions


class BaseNumpyIndexer(BaseVectorIndexer):
    """
    :class:`BaseNumpyIndexer` stores and loads vector in a compresses binary file

    .. note::
        :attr:`compress_level` balances between time and space. By default, :classL`NumpyIndexer` has
        :attr:`compress_level` = 0.

        Setting :attr:`compress_level`>0 gives a smaller file size on the disk in the index time. However, in the query
        time it loads all data into memory at once. Not ideal for large scale application.

        Setting :attr:`compress_level`=0 enables :func:`np.memmap`, which loads data in an on-demand way and
        gives smaller memory footprint in the query time. However, it often gives larger file size on the disk.

    :param compress_level: The compresslevel argument is an integer from 0 to 9 controlling the
                    level of compression; 1 is fastest and produces the least compression,
                    and 9 is slowest and produces the most compression. 0 is no compression
                    at all. The default is 9.
    :param ref_indexer: Bootstrap the current indexer from a ``ref_indexer``. This enables user to switch
                        the query algorithm at the query time.
    :param delete_on_dump: whether to delete the rows marked as delete (see ``valid_indices``)
    """

    def __init__(
        self,
        compress_level: int = 1,
        ref_indexer: Optional['BaseNumpyIndexer'] = None,
        delete_on_dump: bool = False,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.num_dim = None
        self.dtype = None
        self.delete_on_dump = delete_on_dump
        self.compress_level = compress_level
        self.key_bytes = b''
        self.valid_indices = np.array([], dtype=bool)
        self.ref_indexer_workspace_name = None

        if ref_indexer:
            # copy the header info of the binary file
            self.num_dim = ref_indexer.num_dim
            self.dtype = ref_indexer.dtype
            self.compress_level = ref_indexer.compress_level
            self.key_bytes = ref_indexer.key_bytes
            self.key_length = ref_indexer.key_length
            self._size = ref_indexer._size
            # point to the ref_indexer.index_filename
            # so that later in `post_init()` it will load from the referred index_filename
            self.valid_indices = ref_indexer.valid_indices
            self.index_filename = ref_indexer.index_filename
            self.logger.warning(
                f'\n'
                f'num_dim extracted from `ref_indexer` to {ref_indexer.num_dim} \n'
                f'_size extracted from `ref_indexer` to {ref_indexer._size} \n'
                f'dtype extracted from `ref_indexer` to {ref_indexer.dtype} \n'
                f'compress_level overridden from `ref_indexer` to {ref_indexer.compress_level} \n'
                f'index_filename overridden from `ref_indexer` to {ref_indexer.index_filename}'
            )
            self.ref_indexer_workspace_name = ref_indexer.workspace_name
            self.delete_on_dump = getattr(ref_indexer, 'delete_on_dump', delete_on_dump)

    def _delete_invalid_indices(self):
        valid = self.valid_indices[self.valid_indices == True]  # noqa
        if len(valid) != len(self.valid_indices):
            self._clean_memmap()
            self._post_clean_memmap(valid)

    def _post_clean_memmap(self, valid):
        # here we need to make sure the fields
        # that depend on the valid_indices are cleaned up too
        valid_key_bytes = np.frombuffer(
            self.key_bytes, dtype=(np.str_, self.key_length)
        )[self.valid_indices].tobytes()
        self.key_bytes = valid_key_bytes
        self._size = len(valid)
        self.valid_indices = valid
        del self._int2ext_id
        del self._ext2int_id

    def _clean_memmap(self):
        # clean up the underlying matrix of entries marked for deletion
        # first we need to make sure we flush the writing handler
        if self.write_handler and not self.write_handler.closed:
            with self.write_handler as f:
                f.flush()
            self.handler_mutex = False
        # force the raw_ndarray (underlying matrix) to re-read from disk
        # (needed when there were writing ops to be flushed)
        del self._raw_ndarray
        filtered = self._raw_ndarray[self.valid_indices]
        # we need an intermediary file
        tmp_path = self.index_abspath + 'tmp'

        # write the bytes in the respective files
        if self.compress_level > 0:
            with gzip.open(
                tmp_path, 'wb', compresslevel=self.compress_level
            ) as new_gzip_fh:
                new_gzip_fh.write(filtered.tobytes())
        else:
            with open(tmp_path, 'wb') as filtered_data_fh:
                filtered_data_fh.write(filtered.tobytes())

        os.remove(self.index_abspath)
        os.rename(tmp_path, self.index_abspath)
        # force it to re-read again from the new file
        del self._raw_ndarray

    def __getstate__(self):
        # called on pickle save
        if self.delete_on_dump:
            self._delete_invalid_indices()
        d = super().__getstate__()
        return d

    @property
    def workspace_name(self):
        """Get the workspace name.


        .. # noqa: DAR201
        """
        return (
            self.name
            if self.ref_indexer_workspace_name is None
            else self.ref_indexer_workspace_name
        )

    @property
    def index_abspath(self) -> str:
        """Get the file path of the index storage

        Use index_abspath


        .. # noqa: DAR201
        """
        return self.get_file_from_workspace(self.index_filename)

    def get_add_handler(self) -> 'io.BufferedWriter':
        """Open a binary gzip file for appending new vectors

        :return: a gzip file stream
        """
        if self.compress_level > 0:
            return gzip.open(
                self.index_abspath, 'ab', compresslevel=self.compress_level
            )
        else:
            return open(self.index_abspath, 'ab')

    def get_create_handler(self) -> 'io.BufferedWriter':
        """Create a new gzip file for adding new vectors. The old vectors are replaced.

        :return: a gzip file stream
        """
        if self.compress_level > 0:
            return gzip.open(
                self.index_abspath, 'wb', compresslevel=self.compress_level
            )
        else:
            return open(self.index_abspath, 'wb')

    def _validate_key_vector_shapes(self, keys, vectors):
        if len(vectors.shape) != 2:
            raise ValueError(
                f'vectors shape {vectors.shape} is not valid, expecting "vectors" to have rank of 2'
            )

        if not getattr(self, 'num_dim', None):
            self.num_dim = vectors.shape[1]
            self.dtype = vectors.dtype.name
        elif self.num_dim != vectors.shape[1]:
            raise ValueError(
                f'vectors shape {vectors.shape} does not match with indexers\'s dim: {self.num_dim}'
            )
        elif self.dtype != vectors.dtype.name:
            raise TypeError(
                f'vectors\' dtype {vectors.dtype.name} does not match with indexers\'s dtype: {self.dtype}'
            )

        if keys.shape[0] != vectors.shape[0]:
            raise ValueError(
                f'number of key {keys.shape[0]} not equal to number of vectors {vectors.shape[0]}'
            )

    def add(self, keys: Iterable[str], vectors: 'np.ndarray', *args, **kwargs) -> None:
        """Add the embeddings and document ids to the index.

        :param keys: a list of ``id``, i.e. ``doc.id`` in protobuf
        :param vectors: embeddings
        :param args: not used
        :param kwargs: not used
        """
        np_keys = np.array(keys, (np.str_, self.key_length))
        self._add(np_keys, vectors)

    def _add(self, keys: 'np.ndarray', vectors: 'np.ndarray'):
        if keys.size and vectors.size:
            self._validate_key_vector_shapes(keys, vectors)
            self.write_handler.write(vectors.tobytes())
            self.valid_indices = np.concatenate(
                (self.valid_indices, np.full(len(keys), True))
            )
            self.key_bytes += keys.tobytes()
            self._size += keys.shape[0]

    def update(
        self, keys: Iterable[str], vectors: 'np.ndarray', *args, **kwargs
    ) -> None:
        """Update the embeddings on the index via document ids.

        :param keys: a list of ``id``, i.e. ``doc.id`` in protobuf
        :param vectors: embeddings
        :param args: not used
        :param kwargs: not used
        """
        # noinspection PyTypeChecker
        if self.size:
            keys, values = self._filter_nonexistent_keys_values(
                keys, vectors, self._ext2int_id.keys()
            )
            if keys:
                np_keys = np.array(keys, (np.str_, self.key_length))
                self._delete(np_keys)
                self._add(np_keys, np.array(values))
        else:
            self.logger.error(f'{self!r} is empty, update is aborted')

    def _delete(self, keys):
        if keys.size:
            for key in keys:
                # mark as `False` in mask
                self.valid_indices[self._ext2int_id[key]] = False
                self._size -= 1

    def delete(self, keys: Iterable[str], *args, **kwargs) -> None:
        """Delete the embeddings from the index via document ids.

        :param keys: a list of ``id``, i.e. ``doc.id`` in protobuf
        :param args: not used
        :param kwargs: not used
        """
        if self.size:
            keys = self._filter_nonexistent_keys(keys, self._ext2int_id.keys())
            if keys:
                np_keys = np.array(keys, (np.str_, self.key_length))
                self._delete(np_keys)
        else:
            self.logger.error(f'{self!r} is empty, deletion is aborted')

    def get_query_handler(self) -> Optional['np.ndarray']:
        """Open a gzip file and load it as a numpy ndarray

        :return: a numpy ndarray of vectors
        """
        if np.all(self.valid_indices):
            vecs = self._raw_ndarray
        else:
            vecs = self._raw_ndarray[self.valid_indices]

        if vecs is not None:
            return self.build_advanced_index(vecs)

    def build_advanced_index(self, vecs: 'np.ndarray'):
        """Not implemented here.


        .. # noqa: DAR201


        .. # noqa: DAR101
        """
        raise NotImplementedError

    def _load_gzip(self, abspath: str, mode='rb') -> Optional['np.ndarray']:
        try:
            self.logger.info(f'loading index from {abspath}...')
            with gzip.open(abspath, mode) as fp:
                return np.frombuffer(fp.read(), dtype=self.dtype).reshape(
                    [-1, self.num_dim]
                )
        except EOFError:
            self.logger.error(
                f'{abspath} is broken/incomplete, perhaps forgot to ".close()" in the last usage?'
            )

    @cached_property
    def _raw_ndarray(self) -> Optional['np.ndarray']:
        if not (path.exists(self.index_abspath) or self.num_dim or self.dtype):
            return

        if self.compress_level > 0:
            return self._load_gzip(self.index_abspath)
        elif self.size is not None and os.stat(self.index_abspath).st_size:
            self.logger.success(f'memmap is enabled for {self.index_abspath}')
            # `==` is required. `is False` does not work in np
            deleted_keys = len(self.valid_indices[self.valid_indices == False])  # noqa
            return np.memmap(
                self.index_abspath,
                dtype=self.dtype,
                mode='r',
                shape=(self.size + deleted_keys, self.num_dim),
            )

    def query_by_key(
        self, keys: Iterable[str], *args, **kwargs
    ) -> Optional['np.ndarray']:
        """
        Search the index by the external key (passed during `.add(`).

        :param keys: a list of ``id``, i.e. ``doc.id`` in protobuf
        :param args: not used
        :param kwargs: not used
        :return: ndarray of vectors
        """
        keys = self._filter_nonexistent_keys(keys, self._ext2int_id.keys())
        if keys:
            indices = [self._ext2int_id[key] for key in keys]
            return self._raw_ndarray[indices]
        else:
            return None

    @cached_property
    def _int2ext_id(self) -> Optional['np.ndarray']:
        """Convert internal ids (0,1,2,3,4,...) to external ids (random strings)


        .. # noqa: DAR201
        """
        if self.key_bytes:
            r = np.frombuffer(self.key_bytes, dtype=(np.str_, self.key_length))
            # `==` is required. `is False` does not work in np
            deleted_keys = len(self.valid_indices[self.valid_indices == False])  # noqa
            if r.shape[0] == (self.size + deleted_keys) == self._raw_ndarray.shape[0]:
                return r
            else:
                print(
                    f'the size of the keys and vectors are inconsistent '
                    f'({r.shape[0]}, {self._size}, {self._raw_ndarray.shape[0]}), '
                    f'did you write to this index twice? or did you forget to save indexer?'
                )
                self.logger.error(
                    f'the size of the keys and vectors are inconsistent '
                    f'({r.shape[0]}, {self._size}, {self._raw_ndarray.shape[0]}), '
                    f'did you write to this index twice? or did you forget to save indexer?'
                )

    @cached_property
    def _ext2int_id(self) -> Optional[Dict]:
        """Convert external ids (random strings) to internal ids (0,1,2,3,4,...)


        .. # noqa: DAR201
        """
        if self._int2ext_id is not None:
            return {k: idx for idx, k in enumerate(self._int2ext_id)}


@lru_cache(maxsize=3)
def _get_ones(x, y):
    return np.ones((x, y))


def _ext_A(A):
    nA, dim = A.shape
    A_ext = _get_ones(nA, dim * 3)
    A_ext[:, dim : 2 * dim] = A
    A_ext[:, 2 * dim :] = A ** 2
    return A_ext


def _ext_B(B):
    nB, dim = B.shape
    B_ext = _get_ones(dim * 3, nB)
    B_ext[:dim] = (B ** 2).T
    B_ext[dim : 2 * dim] = -2.0 * B.T
    del B
    return B_ext


def _euclidean(A_ext, B_ext):
    sqdist = A_ext.dot(B_ext).clip(min=0)
    return np.sqrt(sqdist)


def _norm(A):
    return A / np.linalg.norm(A, ord=2, axis=1, keepdims=True)


def _cosine(A_norm_ext, B_norm_ext):
    return A_norm_ext.dot(B_norm_ext).clip(min=0) / 2


class NumpyIndexer(BaseNumpyIndexer):
    """An exhaustive vector indexers implemented with numpy and scipy.

    .. note::
        Metrics other than `cosine` and `euclidean` requires ``scipy`` installed.

    :param metric: The distance metric to use. `braycurtis`, `canberra`, `chebyshev`, `cityblock`, `correlation`,
                    `cosine`, `dice`, `euclidean`, `hamming`, `jaccard`, `jensenshannon`, `kulsinski`,
                    `mahalanobis`,
                    `matching`, `minkowski`, `rogerstanimoto`, `russellrao`, `seuclidean`, `sokalmichener`,
                    `sokalsneath`, `sqeuclidean`, `wminkowski`, `yule`.
    :param backend: `numpy` or `scipy`, `numpy` only supports `euclidean` and `cosine` distance
    :param compress_level: compression level to use
    """

    batch_size = 512

    def __init__(
        self,
        metric: str = 'cosine',
        backend: str = 'numpy',
        compress_level: int = 0,
        *args,
        **kwargs,
    ):
        super().__init__(*args, compress_level=compress_level, **kwargs)
        self.metric = metric
        self.backend = backend

    @staticmethod
    def _get_sorted_top_k(
        dist: 'np.array', top_k: int
    ) -> Tuple['np.ndarray', 'np.ndarray']:
        """Find top-k smallest distances in ascending order.

        Idea is to use partial sort to retrieve top-k smallest distances unsorted and then sort these
        in ascending order. Equivalent to full sort but faster for n >> k. If k >= n revert to full sort.

        :param dist: the distances
        :param top_k: nr to limit
        :return: tuple of indices, computed distances
        """
        if top_k >= dist.shape[1]:
            idx = dist.argsort(axis=1)[:, :top_k]
            dist = np.take_along_axis(dist, idx, axis=1)
        else:
            idx_ps = dist.argpartition(kth=top_k, axis=1)[:, :top_k]
            dist = np.take_along_axis(dist, idx_ps, axis=1)
            idx_fs = dist.argsort(axis=1)
            idx = np.take_along_axis(idx_ps, idx_fs, axis=1)
            dist = np.take_along_axis(dist, idx_fs, axis=1)

        return idx, dist

    def query(
        self, vectors: 'np.ndarray', top_k: int, *args, **kwargs
    ) -> Tuple['np.ndarray', 'np.ndarray']:
        """Find the top-k vectors with smallest ``metric`` and return their ids in ascending order.

        :return: a tuple of two ndarray.
            The first is ids in shape B x K (`dtype=int`), the second is metric in shape B x K (`dtype=float`)

        .. warning::
            This operation is memory-consuming.

            Distance (the smaller the better) is returned, not the score.

        :param vectors: the vectors with which to search
        :param args: not used
        :param kwargs: not used
        :param top_k: nr of results to return
        :return: tuple of indices within matrix and distances
        """
        if self.size == 0:
            return np.array([]), np.array([])
        if self.metric not in {'cosine', 'euclidean'} or self.backend == 'scipy':
            dist = self._cdist(vectors, self.query_handler)
        elif self.metric == 'euclidean':
            _query_vectors = _ext_A(vectors)
            dist = self._euclidean(_query_vectors, self.query_handler)
        elif self.metric == 'cosine':
            _query_vectors = _ext_A(_norm(vectors))
            dist = self._cosine(_query_vectors, self.query_handler)

        idx, dist = self._get_sorted_top_k(dist, top_k)
        indices = self._int2ext_id[self.valid_indices][idx]
        return indices, dist

    def build_advanced_index(self, vecs: 'np.ndarray') -> 'np.ndarray':
        """
        Build advanced index structure based on in-memory numpy ndarray, e.g. graph, tree, etc.

        :param vecs: The raw numpy ndarray.
        :return: Advanced index.
        """
        return vecs

    @batching(merge_over_axis=1, slice_on=2)
    def _euclidean(self, cached_A, raw_B):
        data = _ext_B(raw_B)
        return _euclidean(cached_A, data)

    @batching(merge_over_axis=1, slice_on=2)
    def _cosine(self, cached_A, raw_B):
        data = _ext_B(_norm(raw_B))
        return _cosine(cached_A, data)

    @batching(merge_over_axis=1, slice_on=2)
    def _cdist(self, *args, **kwargs):
        with ImportExtensions(required=True):
            from scipy.spatial.distance import cdist
        return cdist(*args, **kwargs, metric=self.metric)



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

schema_requests = {
    'Jina::Requests': {
        'description': 'Define how the Executor behaves under network requests.',
        'type': 'object',
        'additionalProperties': False,
        'properties': {
            'use_default': {
                'description': 'If set, then inherit from the default `Driver` settings for this type of Executor.',
                'type': 'boolean',
                'default': False,
            },
            'on': {
                'description': 'Defines how the `Executor` behaves under different types of request',
                'type': 'object',
                'properties': {
                    f'{r_type}Request': {
                        'type': 'object',
                        'properties': {
                            'with': {
                                'type': 'object',
                                'description': 'The common kwargs that all drivers defined under this Request.',
                            },
                            'drivers': {'$ref': f'#/definitions/Jina::Drivers::All'},
                        },
                        'additionalProperties': False,
                        'description': f'Defines how the `Executor` behaves under {r_type} request.',
                    }
                    for r_type in [
                        'Index',
                        'Train',
                        'Search',
                        'Update',
                        'Delete',
                        'Control',
                    ]
                },
                'additionalProperties': False,
            },
        },
    }
}



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

import time

from google.protobuf.json_format import MessageToJson

from . import BaseDriver
from ..types.querylang.queryset.dunderkey import dunder_get
from ..excepts import UnknownControlCommand, RuntimeTerminated, NoExplicitMessage
from ..proto import jina_pb2


class BaseControlDriver(BaseDriver):
    """Control driver does not have access to the executor and it
    often works directly with protobuf layer instead Jina primitive types"""

    @property
    def envelope(self) -> 'jina_pb2.EnvelopeProto':
        """Get the current request, shortcut to ``self.runtime.message``


        .. # noqa: DAR201
        """
        return self.msg.envelope


class LogInfoDriver(BaseControlDriver):
    """
    Log output the request info

    :param key: (str) that represents a first level or nested key in the dict
    :param json: (bool) indicating if the log output should be formatted as json
    :param args: additional positional arguments which are just used for the parent initialization
    :param kwargs: additional key value arguments which are just used for the parent initialization
    """

    def __init__(self, key: str = 'request', json: bool = True, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.key = key
        self.json = json

    def __call__(self, *args, **kwargs):
        """Log the information.

        :param args: unused
        :param kwargs: unused
        """
        data = dunder_get(self.msg.proto, self.key)
        if self.json:
            self.logger.info(MessageToJson(data))
        else:
            self.logger.info(data)


class WaitDriver(BaseControlDriver):
    """Wait for some seconds, mainly for demo purpose"""

    def __call__(self, *args, **kwargs):
        """Wait for some seconds, mainly for demo purpose


        .. # noqa: DAR101
        """
        time.sleep(5)


class ControlReqDriver(BaseControlDriver):
    """Handling the control request, by default it is installed for all :class:`jina.peapods.peas.BasePea`"""

    def __call__(self, *args, **kwargs):
        """Handle the request controlling.

        :param args: unused
        :param kwargs: unused
        """
        if self.req.command == 'TERMINATE':
            self.envelope.status.code = jina_pb2.StatusProto.SUCCESS
            raise RuntimeTerminated
        elif self.req.command == 'STATUS':
            self.envelope.status.code = jina_pb2.StatusProto.READY
            self.req.args = vars(self.runtime.args)
        else:
            raise UnknownControlCommand(f'don\'t know how to handle {self.req.command}')


class RouteDriver(ControlReqDriver):
    """Ensures that data requests are forwarded to the downstream `:class:`BasePea` ensuring
      that the load is balanced between parallel `:class:`BasePea` if the scheduling `:class:`SchedulerType` is LOAD_BALANCE.

    .. note::
        - The dealer never receives a control request from the router,
        every time it finishes a job and sends via out_sock, it returns the envelope with control
        request idle back to the router. The dealer also sends control request idle to the router
        when it first starts.

        - The router receives requests from both dealer and upstream pusher.
         if it is an upstream request, use LB to schedule the receiver,
         mark it in the envelope if it is a control request in

    :param raise_no_dealer: raise a RuntimeError when no available dealer
    :param args: additional positional arguments which are just used for the parent initialization
    :param kwargs: additional key value arguments which are just used for the parent initialization
    """

    def __init__(self, raise_no_dealer: bool = False, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.idle_dealer_ids = set()
        self.is_polling_paused = False
        self.raise_no_dealer = raise_no_dealer

    def __call__(self, *args, **kwargs):
        """Perform the routing.

        :param args: additional positional arguments which are just used for calling the parent
        :param kwargs: additional key value arguments which are just used for calling the parent


        .. # noqa: DAR401
        """
        if self.msg.is_data_request:
            self.logger.debug(self.idle_dealer_ids)
            if self.idle_dealer_ids:
                dealer_id = self.idle_dealer_ids.pop()
                self.envelope.receiver_id = dealer_id
                if not self.idle_dealer_ids:
                    self.runtime._zmqlet.pause_pollin()
                    self.is_polling_paused = True
            elif self.raise_no_dealer:
                raise RuntimeError(
                    'if this router connects more than one dealer, '
                    'then this error should never be raised. often when it '
                    'is raised, some Pods must fail to start, so please go '
                    'up and check the first error message in the log'
                )
            # else:
            #    this FALLBACK to trivial message pass
            #
            # Explanation on the logic here:
            # there are two cases that when `idle_dealer_ids` is empty
            # (1) this driver is used in a PUSH-PULL fan-out setting,
            # where no dealer is registered in the first place, so `idle_dealer_ids` is empty
            # all the time
            # (2) this driver is used in a ROUTER-DEALER fan-out setting,
            # where some dealer is broken/fails to start, so `idle_dealer_ids` is empty
        elif self.req.command == 'IDLE':
            self.idle_dealer_ids.add(self.envelope.receiver_id)
            self.logger.debug(
                f'{self.envelope.receiver_id} is idle, now I know these idle peas {self.idle_dealer_ids}'
            )
            if self.is_polling_paused:
                self.runtime._zmqlet.resume_pollin()
                self.is_polling_paused = False
            raise NoExplicitMessage
        else:
            super().__call__(*args, **kwargs)


class ForwardDriver(RouteDriver):
    """Alias to :class:`RouteDriver`"""



import os

import numpy as np

from jina.drivers import FlatRecursiveMixin, BaseRecursiveDriver

if False:
    # noinspection PyUnreachableCode
    from jina import DocumentSet


class PngToDiskDriver(FlatRecursiveMixin, BaseRecursiveDriver):
    """A driver that can store an intermediate representation of a png in the workspace, under a given folder.

    Useful for debugging Crafters in the Flow

    :param workspace: the folder where we store the pngs
    :param prefix: the subfolder to add to workspace
    :param top: limit the pngs to first N
    """

    def __init__(self, workspace, prefix='', top=10, *args, **kwargs):
        self.prefix = prefix
        self.top = top
        self.done = 0
        self.workspace = workspace
        self.folder = os.path.join(self.workspace, self.prefix)
        if not os.path.exists(self.folder):
            os.makedirs(self.folder)
        super().__init__(*args, **kwargs)

    def _apply_all(
        self,
        docs: 'DocumentSet',
        *args,
        **kwargs,
    ) -> None:
        def _move_channel_axis(
            img: 'np.ndarray', channel_axis_to_move: int, target_channel_axis: int = -1
        ) -> 'np.ndarray':
            if channel_axis_to_move == target_channel_axis:
                return img
            return np.moveaxis(img, channel_axis_to_move, target_channel_axis)

        def _load_image(blob: 'np.ndarray', channel_axis: int):
            from PIL import Image

            img = _move_channel_axis(blob, channel_axis)
            return Image.fromarray(img.astype('uint8'))

        for d in docs:
            if self.done < self.top:
                img = _load_image(d.blob, -1)
                path = os.path.join(self.folder, f'{self.done}.png')
                img.save(path)
                self.done += 1



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

import numpy as np

from jina.executors.decorators import batching, as_ndarray
from jina.executors.encoders import BaseVideoEncoder
from jina.executors.encoders.frameworks import BaseTorchEncoder


class VideoTorchEncoder(BaseTorchEncoder, BaseVideoEncoder):
    """
    Encode data from a ndarray, using the models from `torchvision.models`.

    :class:`VideoTorchEncoder` encodes data from a ndarray, potentially
    B x T x (Channel x Height x Width) into an ndarray of `B x D`.
    Internally, :class:`VideoTorchEncoder` wraps the models from
    `torchvision.models`: https://pytorch.org/docs/stable/torchvision/models.html

    :param model_name: the name of the model.
        Supported models include ``r3d_18``, ``mc3_18``, ``r2plus1d_18``
        Default is ``r3d_18``.
    :param pool_strategy: the pooling strategy
        - `None` means that the output of the model will be the 4D tensor
            output of the last convolutional block.
        - `mean` means that global average pooling will be applied to the
            output of the last convolutional block, and thus the output
            of the model will be a 2D tensor.
        - `max` means that global max pooling will be applied.
    """

    def __init__(self,
                 model_name: str = 'r3d_18',
                 channel_axis: int = 1,
                 pool_strategy: str = 'mean',
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.channel_axis = channel_axis
        self.model_name = model_name
        self._default_channel_axis = 2
        if pool_strategy not in ('mean', 'max', None):
            raise NotImplementedError(f'unknown pool_strategy: {self.pool_strategy}')
        self.pool_strategy = pool_strategy

    def post_init(self):
        """Load Torch model"""
        super().post_init()
        import torchvision.models.video as models
        if self.pool_strategy is not None:
            self.pool_fn = getattr(np, self.pool_strategy)
        self.model = getattr(models, self.model_name)(pretrained=True).eval()
        self.to_device(self.model)

    def _get_features(self, x):
        x = self.model.stem(x)
        x = self.model.layer1(x)
        x = self.model.layer2(x)
        x = self.model.layer3(x)
        x = self.model.layer4(x)
        x = self.model.avgpool(x)
        x = x.flatten(1)
        return x

    def _get_pooling(self, feature_map: 'np.ndarray') -> 'np.ndarray':
        if feature_map.ndim == 2 or self.pool_strategy is None:
            return feature_map
        return self.pool_fn(feature_map, axis=(2, 3))

    @batching
    @as_ndarray
    def encode(self, data: 'np.ndarray', *args, **kwargs) -> 'np.ndarray':
        """
         Encode data from a ndarray.

         :param data: a `B x T x (Channel x Height x Width)` numpy ``ndarray``,
            `B` is the size of the batch, `T` is the number of frames
         :return: a `B x D` numpy ``ndarray``, `D` is the output dimension
        """
        if self.channel_axis != self._default_channel_axis:
            data = np.moveaxis(data, self.channel_axis, self._default_channel_axis)
        import torch
        _input = torch.from_numpy(data.astype('float32'))
        if self.on_gpu:
            _input = _input.cuda()
        _feature = self._get_features(_input).detach()
        if not self.on_gpu:
            _feature = _feature.cpu()
        _feature = _feature.numpy()
        return self._get_pooling(_feature)


import inspect
import re
import typing
from functools import reduce


def _python_type_to_schema_type(p):
    if p == 'str':
        dtype = 'string'
    elif p == 'int' or p == 'float':
        dtype = 'number'
    elif p in {'typing.List[str]', 'typing.Tuple[str]', 'list', 'tuple'}:
        dtype = 'array'
    elif p == 'bool':
        dtype = 'boolean'
    elif p == 'dict':
        dtype = 'object'
    else:
        dtype = None
        # raise TypeError(f'{p} is not supported')

    return dtype


def _cli_to_schema(
    api_dict,
    target,
    extras=None,
    required=None,
    allow_addition=False,
    namespace='Jina',
    description='',
):
    pod_api = None

    for d in api_dict['methods']:
        if d['name'] == target:
            pod_api = d['options']
            break

    _schema = {
        'properties': {},
        'type': 'object',
        'required': [],
        'additionalProperties': allow_addition,
        'description': description,
    }

    for p in pod_api:
        dtype = _python_type_to_schema_type(p['type'])
        pv = {'description': p['help'].strip(), 'type': dtype, 'default': p['default']}
        if p['choices']:
            pv['enum'] = p['choices']
        if p['required']:
            _schema['required'].append(p['name'])
        if dtype == 'array':
            _schema['items'] = {'type': 'string', 'minItems': 1, 'uniqueItems': True}

        _schema['properties'][p['name']] = pv

    if extras:
        _schema['properties'].update(extras)
    if required:
        _schema['required'].extend(required)

    return {f'{namespace}::{target.capitalize()}': _schema}


def _get_all_arguments(class_):
    def get_class_arguments(class_):
        """
        :param class_: the class to check
        :return: a list containing the arguments from `class_`
        """
        taboo = {'self', 'args', 'kwargs'}
        signature = inspect.signature(class_.__init__)

        reg = r'.*?:param.*?%s:(.*)'

        class_arguments = {}
        for p in signature.parameters.values():
            if p.name in taboo:
                continue
            class_arguments[p.name] = {}
            if p.default != inspect._empty:
                class_arguments[p.name]['default'] = p.default
            else:
                class_arguments[p.name]['default'] = None
            if p.annotation != inspect._empty:
                dtype = None
                try:
                    if (
                        hasattr(p.annotation, '__origin__')
                        and p.annotation.__origin__ is typing.Union
                    ):
                        dtype = p.annotation.__args__[0].__name__
                    else:
                        dtype = getattr(
                            p.annotation, '__origin__', p.annotation
                        ).__name__
                except:
                    pass
                dtype = _python_type_to_schema_type(dtype)
                if dtype:
                    class_arguments[p.name]['type'] = dtype

            if class_.__init__.__doc__:
                m = re.search(reg % p.name, class_.__init__.__doc__)
                if m and m.group(1):
                    class_arguments[p.name]['description'] = m.group(1).strip()

        return class_arguments

    def accumulate_classes(cls):
        """
        :param cls: the class to check
        :return: all classes from which cls inherits from
        """

        def _accumulate_classes(c, cs):
            cs.append(c)
            if cls == object:
                return cs
            for base in c.__bases__:
                _accumulate_classes(base, cs)
            return cs

        classes = []
        _accumulate_classes(cls, classes)
        return set(classes)

    all_classes = accumulate_classes(class_)
    args = list(map(lambda x: get_class_arguments(x), all_classes))
    return reduce(lambda x, y: {**x, **y}, args)


def _jina_class_to_schema(cls):
    kwargs = _get_all_arguments(cls)

    return {
        'type': 'object',
        'description': cls.__doc__.strip() if cls.__doc__ else '',
        'properties': {
            'jtype': {
                'type': 'string',
                'const': cls.__name__,
                'description': cls.__doc__.strip().split('\n')[0]
                if cls.__doc__
                else '',
            },
            'with': {
                'type': 'object',
                'description': 'The arguments of this Jina Executor/Driver',
                'properties': kwargs,
                'additionalProperties': False,
            },
            'metas': {'$ref': '#/definitions/Jina::Metas'},
            'requests': {'$ref': '#/definitions/Jina::Requests'},
        },
        'additionalProperties': False,
    }
